#!/usr/bin/env python3

import sys
from textwrap import dedent


def main(argv):
    print(dedent(f"""\
    test-{argv[0]}:
      - echo generated test variant
    """), end='')


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (c) 2018 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import logging
from pathlib import Path

import click
from dateutil.tz import tzutc
import git

from configparser import (
        NoOptionError,
        NoSectionError,
    )
from . import utils
from ..errors import VersioningError
from ..git_time import (
    determine_source_date,
    determine_version,
)
from ..versioning import SemVer

log = logging.getLogger(__name__)


@click.pass_context
def initialize_global_variables_from_config(ctx, config):
    assert ctx.obj.workspace
    ctx.obj.config = config
    config_file = set_path_variables(ctx.obj.workspace)
    set_version_variables(config_file)


@click.pass_context
def set_path_variables(ctx, workspace):
    try:
        with git.Repo(workspace) as repo, repo.config_reader() as cfg:
            code_dir = workspace / cfg.get_value('hopic.code', 'dir')
    except (git.InvalidGitRepositoryError, git.NoSuchPathError, NoOptionError, NoSectionError):
        code_dir = workspace

    ctx.obj.code_dir = code_dir
    ctx.obj.volume_vars['WORKSPACE'] = str(code_dir)
    source_date = determine_source_date(code_dir)
    if source_date is not None:
        ctx.obj.source_date = source_date
        ctx.obj.source_date_epoch = int((
            source_date - datetime.utcfromtimestamp(0).replace(tzinfo=tzutc())
        ).total_seconds())
        ctx.obj.volume_vars['SOURCE_DATE_EPOCH'] = str(ctx.obj.source_date_epoch)
    try:
        config_file = utils.determine_config_file_name(ctx, workspace)
    except click.BadParameter:
        return None

    if config_file is not None:
        if not config_file.is_absolute() and not config_file.is_reserved():
            config_file = Path.cwd() / config_file
        ctx.obj.config_dir = config_file.parent
        ctx.obj.volume_vars['CFGDIR'] = str(ctx.obj.config_dir)

    return config_file


@click.pass_context
def set_version_variables(ctx, config_file, *, config=None):
    version_info = config.get('version', {}) if config is not None else ctx.obj.config.get('version', {})

    ctx.obj.version, commit_hash = determine_version(
            version_info,
            config_dir=(config_file and ctx.obj.config_dir),
            code_dir=ctx.obj.code_dir,
        )

    if ctx.obj.version is None:
        if version_info.get('tag'):
            error_msg = (
                "Failed to determine the current version from Git tag. "
                "If this is a new repository, please create a version tag. "
                "If this is a shallow clone, try a deepening fetch."
            )
        elif version_info.get('file'):
            error_msg = (
                "Failed to determine the current version from file. "
                "Make sure the file exists and contains the expected value."
            )
        else:
            error_msg = "Failed to determine the current version."

        ctx.obj.volume_vars['VERSION'] = ctx.obj.volume_vars['PURE_VERSION'] = ctx.obj.volume_vars['PUBLISH_VERSION'] = VersioningError(error_msg)
        return

    log.debug("read version: \x1B[34m%s\x1B[39m", ctx.obj.version)
    ctx.obj.volume_vars['VERSION'] = str(ctx.obj.version)
    ctx.obj.volume_vars["PURE_VERSION"] = str(ctx.obj.version.without_meta())
    # FIXME: make this conversion work even when not using SemVer as versioning policy
    # Convert SemVer to Debian version: '~' for pre-release instead of '-'
    ctx.obj.volume_vars['DEBVERSION'] = ctx.obj.volume_vars['VERSION'].replace('-', '~', 1).replace('.dirty.', '+dirty', 1)
    if ctx.obj.publishable_version:
        ctx.obj.volume_vars['PUBLISH_VERSION'] = ctx.obj.volume_vars['PURE_VERSION']
        if 'build' in version_info:
            ctx.obj.volume_vars['PUBLISH_VERSION'] += f"+{version_info['build']}"
    else:
        assert commit_hash
        ver = SemVer.parse(ctx.obj.volume_vars['VERSION'])
        ver.build = ()  # discard duplicate commit_hash
        ver.prerelease += (commit_hash, )
        if 'build' in version_info:
            ver.build = (version_info['build'],)
        ctx.obj.volume_vars['PUBLISH_VERSION'] = str(ver)



# -*- coding: utf-8 -*-

# Copyright (c) 2019 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import os
import re
import subprocess
import sys
from textwrap import dedent

import git
import pytest

from . import config_file

from .. import credentials
from ..build import HopicGitInfo
from ..cli import utils
from ..errors import VersionBumpMismatchError, VersioningError
from ..template.utils import command

_git_time = f"{42 * 365 * 24 * 3600} +0000"
_author = git.Actor('Bob Tester', 'bob@example.net')
_commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=_author,
        committer=_author,
    )


@pytest.mark.parametrize("variable", (
    "AUTOSQUASHED_COMMIT",
    "AUTOSQUASHED_COMMITS",
    "SOURCE_COMMIT",
    "SOURCE_COMMITS",
))
def test_autosquash_base(capfd, run_hopic, variable):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write(
                dedent(
                    """\
                    version:
                      bump: no

                    phases:
                      build:
                        test:
                    """
                )
            )
            if variable.endswith("S"):
                f.write(
                    dedent(
                        f"""\
                        #
                              - sh: git log --format=%P ${{{variable}}}
                        """
                    )
                )
            else:
                f.write(
                    dedent(
                        f"""\
                        #
                              - foreach: {variable}
                                sh: git log -1 --format=%P ${{{variable}}}
                        """
                    )
                )
        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)

        # Main branch moves on
        with (run_hopic.toprepo / 'A.txt').open('w') as f:
            f.write('A')
        repo.index.add(('A.txt',))
        final_commit = repo.index.commit(message='feat: add A', **_commitargs)

        # PR branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message='feat: add something useful', **_commitargs)

        # A fixup on top of that change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('useful')
        repo.index.add(('something.txt',))
        repo.index.commit(message='fixup! feat: add something useful', **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree', '--author-name', _author.name, '--author-email', _author.email,
                'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful'),
            ('build',),
        )
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    build_out = ''.join(out.splitlines(keepends=True)[2:])
    commits = build_out.split()
    assert str(final_commit) not in commits
    assert str(base_commit) in commits


def hopic_config_subdir_version_file_tester(capfd, config_dir, hopic_config, version_file, version_input, expected_version, run_hopic, expect_tag=True):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        if not os.path.exists(run_hopic.toprepo / config_dir):
            os.mkdir(run_hopic.toprepo / config_dir)
        with (run_hopic.toprepo / config_dir / 'hopic-ci-config.yaml').open('w') as f:
            f.write(hopic_config)

        with (run_hopic.toprepo / config_dir / version_file).open('w') as f:
            f.write(version_input)
        repo.index.add((os.path.join(config_dir, 'hopic-ci-config.yaml'),))
        repo.index.add((os.path.join(config_dir, version_file),))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)

        # PR branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message='feat: add something useful', **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ('--workspace', './', '--config', os.path.join(config_dir, 'hopic-ci-config.yaml'),
         'checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        ('--workspace', './', '--config', os.path.join(config_dir, 'hopic-ci-config.yaml'),
         'prepare-source-tree', '--author-name', _author.name, '--author-email', _author.email,
         'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful'),
        ('--workspace', './', '--config', os.path.join(config_dir, 'hopic-ci-config.yaml'), 'submit'),
    )
    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    _, merge_commit, version_out, *_ = out.splitlines()
    assert version_out == expected_version
    with git.Repo(run_hopic.toprepo, expand_vars=False) as repo:
        repo.git.checkout('master')
        if expect_tag:
            assert repo.git.tag(l=True) == expected_version

        note = repo.git.notes('show', merge_commit, ref='hopic/master')
        assert re.match(
                r'^Committed-by: Hopic.*\nWith Python version: .*\nAnd with these installed packages:\n.*\bhopic\b',
                note, flags=re.DOTALL | re.MULTILINE,
            )

    return run_hopic.toprepo


def test_hopic_config_subdir_version_file(capfd, run_hopic):
    version = "0.0.1-SNAPSHOT"
    commit_version = "0.0.1"
    version_file = "revision_test.txt"
    config_dir = "test_config"
    hopic_config_subdir_version_file_tester(capfd,
                                            config_dir,
                                            f"""\
version:
  file: {version_file}
  tag:  true
  bump: patch
  format: semver""",
                                            version_file,
                                            f"""\
version={version}""",
                                            commit_version,
                                            run_hopic)


def test_hopic_config_subdir_version_file_after_submit(capfd, run_hopic):
    version = "0.0.42-SNAPSHOT"
    commit_version = "0.0.42"
    version_file = "revision_test.txt"
    config_dir = ".ci"
    test_repo = hopic_config_subdir_version_file_tester(capfd,
                                                        config_dir,
                                                        f"""\
version:
  file: {version_file}
  tag:  true
  bump: patch
  format: semver
  after-submit:
    bump: prerelease
    prerelease-seed: PRERELEASE-TEST""",
                                                        version_file,
                                                        f"""\
version={version}""",
                                                        commit_version,
                                                        run_hopic)
    with (test_repo / config_dir / version_file).open('r') as f:
        assert f.read() == "version=0.0.43-PRERELEASE-TEST"


def test_version_bump_after_submit_from_repo_root_dir(capfd, run_hopic):
    version = "0.0.3-SNAPSHOT"
    commit_version = "0.0.3"
    version_file = "revision_test.txt"
    config_dir = ""
    test_repo = hopic_config_subdir_version_file_tester(capfd,
                                                        config_dir,
                                                        f"""\
version:
  file: {version_file}
  tag:  true
  bump: patch
  format: semver
  after-submit:
    bump: prerelease
    prerelease-seed: PRERELEASE-TEST""",
                                                        version_file,
                                                        f"""\
version={version}""",
                                                        commit_version,
                                                        run_hopic)
    with (test_repo / config_dir / version_file).open('r') as f:
        assert f.read() == "version=0.0.4-PRERELEASE-TEST"


def test_version_file_without_tag_and_bump(capfd, run_hopic):
    version = '1.2.3'
    expected_version = version
    version_file = "revision_test.txt"
    config_dir = '.ci'
    hopic_config_subdir_version_file_tester(
        capfd,
        config_dir,
        dedent(f"""\
            version:
              file: {version_file}
              tag:  no
              bump: no
            """),
        version_file,
        dedent(f"""\
            version={version}"""),
        expected_version,
        run_hopic,
        expect_tag=False,
    )


def merge_conventional_bump(capfd, run_hopic, message, strict=False, on_every_change=True, target='master', merge_message=None):
    if merge_message is None:
        merge_message = message
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write('''\
version:
  format: semver
  tag:    true
  bump:
    policy: conventional-commits
''')
            if strict:
                f.write('    strict: yes\n')
            if not on_every_change:
                f.write('    on-every-change: no\n')
        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch(target, move=True)
        repo.create_tag('0.0.0')

        # PR branch
        repo.head.reference = repo.create_head('something-useful')
        assert not repo.head.is_detached

        # A preceding commit on this PR to detect whether we check more than the first commit's message in a PR
        repo.index.commit(message='chore: some intermediate commit', **_commitargs)
        print(repo.git.log(format='fuller', color=True, stat=True), file=sys.stderr)

        # Some change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message=message, **_commitargs)

        # A succeeding commit on this PR to detect whether we check more than the last commit's message in a PR
        repo.index.commit(message='chore: some other intermediate commit', **_commitargs)
        print(repo.git.log(format='fuller', color=True, stat=True), file=sys.stderr)

    # Successful checkout and build
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', target),
            ('prepare-source-tree',
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                '--author-name', _author.name,
                '--author-email', _author.email,
                'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful', '--title', merge_message),
        )
    return result


def test_merge_conventional_refactor_no_bump(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='refactor: some problem')
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('0.0.1-'), "post merge version should be a pre-release of 0.0.1, not 0.0.1 itself"


def test_merge_conventional_fix_bump(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='fix: some problem')
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('0.0.1+g')


def test_merge_conventional_feat_bump(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='feat: add something useful')
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('0.1.0+g')


def test_merge_conventional_breaking_change_bump(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='refactor!: make the API type better')
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('1.0.0+g')


def test_merge_conventional_feat_with_breaking_bump(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='''\
refactor!: add something awesome

This adds the new awesome feature.

BREAKING CHANGE: unfortunately this was incompatible with the old feature for
the same purpose, so you'll have to migrate.
''')
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('1.0.0+g')


def test_merge_conventional_broken_feat(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='feat add something useful', strict=True)
    assert result.exit_code != 0


def test_merge_conventional_feat_bump_not_on_change(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='feat: add something useful', on_every_change=False)
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit, merge_commit, merge_version = out.splitlines()
    assert merge_version.startswith('0.0.1-4+g')


def test_merge_conventional_breaking_change_on_major_branch(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='refactor!: make the API type better', target='release/42')
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert 'Breaking changes are not allowed' in err


def test_merge_conventional_feat_on_minor_branch(capfd, run_hopic):
    result = merge_conventional_bump(capfd, run_hopic, message='feat: add something useful', target='release/42.21')
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert 'New features are not allowed' in err


def test_move_submodule(capfd, monkeypatch, run_hopic, tmp_path):
    old_subcommand_getter = git.cmd.Git.__getattr__

    def new_subcommand_getter(self, name: str):
        if name == "submodule":
            self(c="protocol.file.allow=always")
        return old_subcommand_getter(self, name)

    monkeypatch.setattr("git.cmd.Git.__getattr__", new_subcommand_getter)

    subrepo = tmp_path / 'subrepo'
    with git.Repo.init(str(subrepo), expand_vars=False) as repo:
        with (subrepo / 'dummy.txt').open('w') as f:
            f.write('Lalalala!\n')
        repo.index.add(('dummy.txt',))
        repo.index.commit(message='Initial dummy commit', **_commitargs)

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write('''\
version:
  bump: no

phases:
  build:
    test:
      - cat subrepo_test/dummy.txt
''')
        repo.index.add(('hopic-ci-config.yaml',))
        repo.git.submodule(('add', subrepo, 'subrepo_test'))
        repo.index.add(('.gitmodules',))
        repo.index.commit(message='Initial commit', **_commitargs)

        # Move submodule
        repo.create_head("move_submodule_branch")
        repo.git.checkout("move_submodule_branch")
        repo.index.remove(["subrepo_test"])
        with (run_hopic.toprepo / ".gitmodules").open("r+") as f:
            f.truncate(0)

        repo.git.submodule(("add", subrepo, "moved_subrepo"))
        repo.index.commit(message="Move submodule", **_commitargs)

    (result,) = run_hopic(('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'))
    assert result.exit_code == 0
    assert (run_hopic.toprepo / 'subrepo_test' / 'dummy.txt').is_file()
    assert not (run_hopic.toprepo / 'moved_subrepo' / 'dummy.txt').is_file()

    (result,) = run_hopic(
        ('--workspace', run_hopic.toprepo, 'prepare-source-tree', '--author-name', _author.name, '--author-email', _author.email,
         'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'move_submodule_branch'))
    assert result.exit_code == 0
    assert not (run_hopic.toprepo / 'subrepo_test' / 'dummy.txt').is_file()
    assert (run_hopic.toprepo / 'moved_subrepo' / 'dummy.txt').is_file()

    # Do checkout of master again to fake build retrigger of an PR
    (result,) = run_hopic(('--workspace', run_hopic.toprepo, 'checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'))
    assert result.exit_code == 0
    assert (run_hopic.toprepo / 'subrepo_test' / 'dummy.txt').is_file()
    assert not (run_hopic.toprepo / 'moved_subrepo' / 'dummy.txt').is_file()


def test_modality_merge_has_all_parents(run_hopic, monkeypatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with open(run_hopic.toprepo / 'hopic-ci-config.yaml', 'w') as f:
            f.write(dedent('''\
                version:
                  bump: no

                modality-source-preparation:
                  AUTO_MERGE:
                    - git fetch origin release/0
                    - sh: git merge --no-commit --no-ff FETCH_HEAD
                      changed-files: []
                      commit-message: "Merge branch 'release/0'"
                '''))
        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)

        # Main branch moves on
        with (run_hopic.toprepo / 'A.txt').open('w') as f:
            f.write('A')
        repo.index.add(('A.txt',))
        final_commit = repo.index.commit(message='feat: add A', **_commitargs)

        # release branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head('release/0', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with open(run_hopic.toprepo / 'something.txt', 'w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        merge_commit = repo.index.commit(message='feat: add something useful', **_commitargs)

    monkeypatch.setenv('GIT_COMMITTER_NAME' , 'My Name is Nobody')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'nobody@example.com')
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-name', _author.name,
                '--author-email', _author.email,
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                'apply-modality-change', 'AUTO_MERGE'),
            ('submit',),
        )
    assert result.exit_code == 0

    with git.Repo(run_hopic.toprepo, expand_vars=False) as repo:
        assert repo.heads.master.commit.parents == (final_commit, merge_commit), f"Produced commit {repo.heads.master.commit} is not a merge commit"

        note = repo.git.notes('show', 'master', ref='hopic/master')
        assert re.match(
                r'^Committed-by: Hopic.*\nWith Python version: .*\nAnd with these installed packages:\n.*\bhopic\b',
                note, flags=re.DOTALL | re.MULTILINE,
            )


@pytest.mark.parametrize(
    "msg_prefix, expected_version",
    (
        ("feat: merge", "0.1.0"),
        ("Merge", None),
    ),
)
def test_modality_merge_commit_message(expected_version, msg_prefix, run_hopic, monkeypatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                f"""\
                version:
                  format: semver
                  tag: true
                  bump:
                    policy: conventional-commits
                    strict: yes
                    on-every-change: {json.dumps(expected_version is not None)}

                pass-through-environment-vars:
                  - CUSTOM_VAR

                modality-source-preparation:
                  AUTO_MERGE:
                    - git fetch origin release/0
                    - sh: git merge --no-commit --no-ff FETCH_HEAD
                      changed-files: []
                      commit-message: "{msg_prefix} branch 'release/0': $CUSTOM_VAR"
                """
            )
        )

        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.create_tag('0.0.0')

        # Main branch moves on
        with (run_hopic.toprepo / 'A.txt').open('w') as f:
            f.write('A')
        repo.index.add(('A.txt',))
        repo.index.commit(message='feat: add A', **_commitargs)

        # release branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head('release/0', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with open(run_hopic.toprepo / 'something.txt', 'w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message='feat: add something useful', **_commitargs)

    monkeypatch.setattr(utils, "get_package_version", lambda package: "42.42.42")
    monkeypatch.setenv('GIT_COMMITTER_NAME' , 'My Name is Nobody')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'nobody@example.com')
    monkeypatch.setenv('CUSTOM_VAR', 'custom value')
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-name', _author.name,
                '--author-email', _author.email,
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                'apply-modality-change', 'AUTO_MERGE'),
            ('submit',),
        )

    assert result.exit_code == 0
    with git.Repo(run_hopic.toprepo, expand_vars=False) as repo:
        if expected_version is not None:
            assert repo.git.describe("master") == expected_version

        assert repo.heads.master.commit.message == dedent(
            f"""\
                {msg_prefix} branch 'release/0': custom value

                Merged-by: Hopic 42.42.42
            """
        )


@pytest.mark.parametrize(
    "msg_tag, expected_version",
    (
        ("refactor!", "1.0.0"),
        ("feat", "0.1.0"),
        ("fix", "0.0.1"),
        ("chore", None),
    ),
)
def test_modality_merge_commit_message_dynamic(expected_version, msg_tag, run_hopic, monkeypatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                """\
                version:
                  format: semver
                  tag: true
                  bump:
                    policy: conventional-commits
                    strict: yes
                    on-every-change: yes

                project-name: test-project

                modality-source-preparation:
                  AUTO_MERGE:
                    - git fetch origin release/0
                    - sh: git merge --no-commit --no-ff FETCH_HEAD
                      changed-files: []
                      # Reuse merged commit's commit message tag for the produced merge commit
                      commit-message-cmd:
                        with-credentials:
                          - id: topsecret
                            type: username-password
                            username-variable: MODALITY_AUTHOR
                            password-variable: MODALITY_PASSWORD
                        sh: >
                          sh -c 'git show -q --format=%s MERGE_HEAD | sed "s|: .*|: merge branch '"'"'release/0'"'"'|" && echo && echo "Committed-by: ${MODALITY_AUTHOR}" && echo "Authorized-by: ${MODALITY_PASSWORD}"'
                """
            )
        )

        repo.index.add(("hopic-ci-config.yaml",))
        base_commit = repo.index.commit(message="Initial commit", **_commitargs)
        repo.create_tag("0.0.0")

        # Main branch moves on
        (run_hopic.toprepo / "A.txt").write_text("A")
        repo.index.add(("A.txt",))
        repo.index.commit(message="feat: add A", **_commitargs)

        # release branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head("release/0", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message=f"{msg_tag}: add something useful", **_commitargs)

    username = "Master of the Universe"
    password = "Open Sesame!"

    def get_credential_id(project_name_arg, cred_id):
        assert cred_id == "topsecret"
        assert project_name_arg == "test-project"
        return username, password

    monkeypatch.setattr(credentials, "get_credential_by_id", get_credential_id)
    monkeypatch.setattr(utils, "get_package_version", lambda package: "42.42.42")
    monkeypatch.setenv("GIT_COMMITTER_NAME", "My Name is Nobody")
    monkeypatch.setenv("GIT_COMMITTER_EMAIL", "nobody@example.com")

    (*_, result) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref="master"),
        command(
            "prepare-source-tree",
            author_name=_author.name,
            author_email=_author.email,
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
        )
        + command("apply-modality-change", "AUTO_MERGE"),
        command("submit"),
    )

    assert result.exit_code == 0
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        if expected_version is not None:
            assert repo.git.describe("master") == expected_version

        assert repo.heads.master.commit.message == dedent(
            f"""\
                {msg_tag}: merge branch 'release/0'

                Committed-by: {username}
                Authorized-by: {password}
                Merged-by: Hopic 42.42.42
            """
        )


def test_modality_merge_nop(capfd, run_hopic, monkeypatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                """\
                    modality-source-preparation:
                      AUTO_MERGE:
                        - git fetch origin release/0
                        - sh: git merge --no-commit --no-ff FETCH_HEAD
                          changed-files: []
                          commit-message: "Merge branch 'release/0'"
                """
            )
        )

        repo.index.add(("hopic-ci-config.yaml",))
        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)

        # release branch from just before the main branch's HEAD, with nothing changed on it
        repo.head.reference = repo.create_head("release/0", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

    monkeypatch.setenv("GIT_COMMITTER_NAME", "My Name is Nobody")
    monkeypatch.setenv("GIT_COMMITTER_EMAIL", "nobody@example.com")

    (*_,) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref="master"),
    )

    # just flush
    capfd.readouterr()

    (result,) = run_hopic(
        command(
            "prepare-source-tree",
            author_name=_author.name,
            author_email=_author.email,
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
        )
        + command(
            "apply-modality-change",
            "AUTO_MERGE",
        ),
    )

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert out == "", "'prepare-source-tree apply-modality-change' should give empty stdout when there's nothing to merge"


def test_modality_with_credentials(run_hopic, monkeypatch):
    username = "test_username"
    password = "super_secret"
    credential_id = "test_credentialId"
    project_name = "test-project"

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                f"""\
                    version:
                      bump: no
                    project-name: {project_name}
                    modality-source-preparation:
                      ALPHA:
                        - with-credentials:
                            id: {credential_id}
                          sh: sh -c 'echo -n "$USERNAME:$PASSWORD" > creds.txt'
                          changed-files:
                            - creds.txt
                          commit-message: "chore: embed secret"
                """
            )
        )
        repo.index.add(("hopic-ci-config.yaml",))
        repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.git.branch("master", move=True)

        # switch branch to allow 'master' to get updated
        repo.head.reference = repo.create_head("something-useless")
        assert not repo.head.is_detached

        def get_credential_id(project_name_arg, cred_id):
            assert credential_id == cred_id
            assert project_name == project_name_arg
            return username, password

        monkeypatch.setattr(credentials, "get_credential_by_id", get_credential_id)

        (*_, result) = run_hopic(
            command(
                "checkout-source-tree",
                target_remote=run_hopic.toprepo,
                target_ref="master",
            ),
            command(
                "prepare-source-tree",
                author_date=f"@{_git_time}",
                commit_date=f"@{_git_time}",
                author_name=_author.name,
                author_email=_author.email,
            )
            + command(
                "apply-modality-change",
                "ALPHA",
            ),
            command("submit"),
        )

        assert result.exit_code == 0

        repo.head.reference = repo.branches["master"]
        repo.head.reset(index=True, working_tree=True)

    creds_content = (run_hopic.toprepo / "creds.txt").read_text()
    assert creds_content == f"{username}:{password}"


@pytest.mark.parametrize('modality_message, expected_version', (
    ('feat: some feature', '0.1.0'),
    ('chore: non bumping', '0.0.0'),
))
def test_modality_version_bump(run_hopic, monkeypatch, modality_message, expected_version):
    monkeypatch.setenv('GIT_COMMITTER_NAME' , 'My Name is Nobody')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'nobody@example.com')
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with open(run_hopic.toprepo / 'hopic-ci-config.yaml', 'w') as f:
            f.write(dedent(f"""\
                version:
                  format: semver
                  tag: true
                  bump:
                    policy: conventional-commits
                    strict: yes

                modality-source-preparation:
                  INTAKE:
                    - sh: touch test.txt
                      changed-files: test.txt
                      commit-message: "{modality_message}"
                """))

        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.create_tag('0.0.0', message='first version')
        repo.head.reference = repo.create_head('release/0', base_commit)

    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-name', _author.name,
                '--author-email', _author.email,
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                'apply-modality-change', 'INTAKE'),
            ('submit', '--target-remote', run_hopic.toprepo)
        )

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        repo.git.checkout('master')
        assert repo.git.describe().startswith(expected_version)

    assert result.exit_code == 0


def test_modality_separate_changed_files(run_hopic, monkeypatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                """\
                version:
                  bump: no

                modality-source-preparation:
                  ALPHA:
                    - sh: touch test.txt
                    - changed-files: test.txt
                      commit-message: "chore: ensure file exists"
                """
            )
        )

        repo.index.add(("hopic-ci-config.yaml",))
        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.head.reference = repo.create_head("release/0", base_commit)

    (*_, result) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref="master"),
        command(
            "prepare-source-tree",
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
            author_name=_author.name,
            author_email=_author.email,
        )
        + command(
            "apply-modality-change",
            "ALPHA",
        ),
    )
    assert result.exit_code == 0
    assert re.match(r"^(?:[A-Fa-f0-9]{40}|[A-Fa-f0-9]{64})\n", result.stdout)


@pytest.mark.parametrize('strict, commit_message, merge_message, expected_result', (
    (True , 'feat: some feature', 'feat: some feature', {'version': '0.1.0'}),
    (False, 'chore: non bumping', 'feat: some feature', {'version': '0.1.0'}),
    (True , 'chore: non bumping', 'feat: some feature', {'error'  : VersionBumpMismatchError}),
    (False, 'fix: some fix'     , 'ci: non bumping'   , {'version': '0.0.1'}),
))
def test_merge_change_request_version_bump(capfd, monkeypatch, run_hopic, strict, commit_message, merge_message, expected_result):
    monkeypatch.setenv('GIT_COMMITTER_NAME' , 'My Name is Nobody')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'nobody@example.com')
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write(
                dedent(f"""\
                    version:
                        format: semver
                        tag: true
                        bump:
                            policy: conventional-commits
                            strict: {strict}
                """)
            )
        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.create_tag('0.0.0', message='first version')

        # PR branch from just before the main branch's HEAD
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message=commit_message, **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree', '--author-name', _author.name, '--author-email', _author.email,
                'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful',
                '--title', merge_message),
            ('submit',),
        )

    if 'error' in expected_result:
        assert result.exception is not None
        if not isinstance(result.exception, VersionBumpMismatchError):
            raise result.exception
    else:
        assert result.exit_code == 0
        with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
            repo.git.checkout('master')
            assert repo.git.describe().startswith(expected_result['version'])


def test_separate_modality_change(run_hopic):
    """It should be possible to apply modality changes without requiring to perform a checkout-source-tree first.

    This will allow using this command locally by users and developers to make testing of those configs easier."""

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with open(run_hopic.toprepo / 'hopic-ci-config.yaml', 'w') as f:
            f.write(dedent('''\
                version:
                  bump: no

                modality-source-preparation:
                  CHANGE:
                    - sh: touch new-file.txt
                      changed-files:
                        - new-file.txt
                      commit-message: Add new file
                '''))
        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)

    (result,) = run_hopic(
            ('--workspace', run_hopic.toprepo,
                'prepare-source-tree',
                '--author-name', _author.name,
                '--author-email', _author.email,
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                'apply-modality-change', 'CHANGE'),
        )
    assert result.exit_code == 0

    with git.Repo(run_hopic.toprepo, expand_vars=False) as repo:
        assert repo.head.commit != base_commit
        assert repo.head.commit.parents == (base_commit,)


def test_bundle_prepare_source_tree(run_hopic, tmp_path):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        src = run_hopic.toprepo / "widget.h"
        src.write_text(
            dedent(
                """\
                    #pragma once
                    extern int sqrt(int);
                """
            )
        )

        repo.index.add(("widget.h",))
        repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag("1.0.0")

        with src.open("a") as f:
            f.write("extern float sqrt(float);\n")

        repo.index.add(("widget.h",))
        repo.index.commit(message="feat: support float too", **_commitargs)
        repo.create_tag("1.1.0")
        repo.git.branch("data")

        # throw away HEAD to allow creating a new "initial commit" for Hopic config repo
        repo.git.update_ref(d="HEAD")
        repo.head.reset(index=True, working_tree=True)

        hopic_cfg = run_hopic.toprepo / "hopic-ci-config.yaml"
        hopic_cfg.write_text(
            dedent(
                """\
                    version:
                      file: version.txt
                      tag:  true
                      bump: minor
                      format: semver
                      after-submit:
                        bump: prerelease
                        prerelease-seed: SNAPSHOT

                    scm:
                      git:
                        worktrees:
                          output/folder: data
                        ref: 1.0.0
                    phases:
                      a:
                        x:
                          - worktrees:
                              output/folder:
                                commit-message: "Update documentation for ${VERSION}"
                            sh: ":"
                """
            )
        )
        (run_hopic.toprepo / "version.txt").write_text(
            dedent(
                """\
                blabla
                version=1.0.0
                mooh
                """
            )
        )
        repo.index.add(("hopic-ci-config.yaml", "version.txt"))
        repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.git.branch("master", move=True)

        # PR branch
        repo.head.reference = repo.create_head("something-useful")
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        hopic_cfg.write_text(hopic_cfg.read_text().replace("1.0.0", "1.1.0"))
        repo.index.add(("hopic-ci-config.yaml",))
        repo.index.commit(message="feat: get new float widget", **_commitargs)

        # A fixup on top of that change
        repo.index.commit(message="fixup! feat: get new float widget", **_commitargs)

    transfer_bundle = tmp_path / "transfer.bundle"
    orig_rundir = tmp_path / "rundir-orig"
    (*_, result) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref="master"),
        command(
            "prepare-source-tree",
            author_name=_author.name,
            author_email=_author.email,
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
            bundle=transfer_bundle,
        )
        + command(
            "merge-change-request",
            source_remote=run_hopic.toprepo,
            source_ref="something-useful",
            change_request="42",
            title="feat: get new float widget",
        ),
        rundir=orig_rundir,
    )
    assert result.exit_code == 0

    unbundle_rundir = tmp_path / "rundir-unbundle"
    (*_, result) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref="master"),
        command("unbundle", transfer_bundle),
        command("build"),
        rundir=unbundle_rundir,
    )
    assert result.exit_code == 0

    orig_info = HopicGitInfo.from_repo(orig_rundir)
    unbundle_info = HopicGitInfo.from_repo(unbundle_rundir)

    # Compare all info _except_ refspecs (which differs due to worktrees)
    orig_dict = orig_info._asdict()
    del orig_dict["refspecs"]
    unbundle_dict = unbundle_info._asdict()
    del unbundle_dict["refspecs"]
    assert orig_dict == unbundle_dict

    # Compare everything except for the worktree refspec
    assert orig_info.refspecs == tuple(x for x in unbundle_info.refspecs if not x.endswith(":data"))


@pytest.mark.parametrize('run_on_change, commit_message, expected_version, expect_publish', (
    ('only'            , None             , '0.0.0'  , False),
    ('only'            , 'feat: something', '0.1.0'  , True ),
    ('never'           , None             , '0.0.0'  , True ),
    ('never'           , 'feat: something', '0.1.0'  , False),
    ('always'          , None             , '0.0.0'  , True ),
    ('always'          , 'feat: something', '0.1.0'  , True ),
    ('new-version-only', None             , '0.0.0'  , False),
    ('new-version-only', 'docs: something', '0.0.1-2', False),
    ('new-version-only', 'feat: something', '0.1.0'  , True ),
))
def test_run_on_change(monkeypatch, run_hopic, run_on_change, commit_message, expected_version, expect_publish):
    expected = [
        ('echo', 'build-a', expected_version),
    ]
    if expect_publish:
        expected.append(('echo', 'publish-a', expected_version))

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = 'hopic-ci-config.yaml'

        with (run_hopic.toprepo / cfg_file).open('w') as f:
            f.write(dedent(f"""\
                    version:
                      format: semver
                      tag:    true
                      bump:
                        policy: conventional-commits

                    phases:
                      build:
                        a:
                          - echo build-a ${{PURE_VERSION}}

                      publish:
                        a:
                          - run-on-change: {run_on_change}
                          - echo publish-a ${{PURE_VERSION}}
                    """))

        repo.index.add((cfg_file,))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag('0.0.0')

        # PR branch
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        if commit_message is not None:
            with (run_hopic.toprepo / 'something.txt').open('w') as f:
                f.write('usable')
            repo.index.add(('something.txt',))
            repo.index.commit(message=commit_message, **_commitargs)

    # Successful checkout and build
    cmds = (
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        ) + ((
            ('prepare-source-tree',
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                '--author-name', _author.name,
                '--author-email', _author.email,
                'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful', '--title', commit_message),
        ) if commit_message is not None else ()) + (
            ('build',),
        )
    (*_, result) = run_hopic(*cmds)

    assert result.exit_code == 0
    assert not expected


@pytest.mark.parametrize('init_version, submittable_version, version_build', (
    ('0.0.0', False, None    ),
    ('0.0.0', True , None    ),
    ('0.0.0', False, '1.70.0'),
    ('0.0.0', True , '1.70.0'),
))
def test_run_publish_version(monkeypatch, run_hopic, init_version, submittable_version, version_build):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = 'hopic-ci-config.yaml'

        with (run_hopic.toprepo / cfg_file).open('w') as f:
            f.write(dedent(f"""\
                    version:
                      format: semver
                      tag:    true
                      bump:
                        policy: conventional-commits
                    {('  build: ' + version_build) if version_build else ''}

                    phases:
                      build:
                        a:
                          - echo build-a ${{PURE_VERSION}}
                          - echo build-a ${{PUBLISH_VERSION}}
                    """))

        repo.index.add((cfg_file,))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag(init_version)

        # PR branch
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

    expected_publish_version = init_version
    if not submittable_version:
        expected_publish_version += f"-{str(base_commit)[0:14]}"
    if version_build:
        expected_publish_version += f"+{version_build}"

    expected = [
        ('echo', 'build-a', init_version),
        ('echo', 'build-a', expected_publish_version),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    # Successful checkout and build
    cmds = (
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
        ) + (
            ('--publishable-version', 'build') if submittable_version else
            ('build',),
        )
    (*_, result) = run_hopic(*cmds)

    assert result.exit_code == 0
    assert not expected


@pytest.mark.parametrize('commit_message, expected_version', (
    ('feat: initial test feature', '0.1.0'),
    ('chore: initial test feature', None),
))
def test_post_submit(run_hopic, capfd, monkeypatch, commit_message, expected_version):
    username = 'test_username'
    password = 'super_secret'
    credential_id = 'test_credentialId'
    project_name = 'test-project'
    init_version = '0.0.0'

    expected_post_submit_commands = [
        ('echo', f"{username} {password}"),
    ]
    if expected_version:
        expected_post_submit_commands.append(('echo', 'on new version only'),)

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write(dedent(f'''\
                    project-name: {project_name}
                    version:
                      format: semver
                      tag:    true
                      bump:
                        policy: conventional-commits
                        strict: yes

                    phases:
                      phase:
                        variant:
                          - echo "BUILD VERSION $VERSION"
                      publish:
                        variant:
                          - run-on-change: 'new-version-only'
                          - echo publish-a ${{PURE_VERSION}}

                    post-submit:
                      credential-step:
                        - with-credentials:
                            id: {credential_id}
                        - echo "$USERNAME $PASSWORD"
                      new-version-only-step:
                        - run-on-change: 'new-version-only'
                          sh: echo "on new version only"
                    '''))
        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='chore: initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag(init_version)

        # PR branch
        repo.head.reference = repo.create_head('something-useful')
        assert not repo.head.is_detached

        # Some change
        with (run_hopic.toprepo / 'something.txt').open('w') as f:
            f.write('some text')
        repo.index.add(('something.txt',))
        repo.index.commit(message=commit_message, **_commitargs)

        repo.git.checkout('something-useful')
        assert not repo.head.is_detached

        def get_credential_id(project_name_arg, cred_id):
            assert credential_id == cred_id
            assert project_name == project_name_arg
            return username, password

        monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

        def prepare_subprocess_mock():
            def mock_check_call(args, *popenargs, **kwargs):
                assert tuple(args) == expected_post_submit_commands.pop(0)

            monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

        (*_, hopic_result) = run_hopic(
            ('checkout-source-tree',
             '--target-remote', run_hopic.toprepo,
             '--target-ref', 'master',),
            ('prepare-source-tree',
             '--author-date', f"@{_git_time}",
             '--commit-date', f"@{_git_time}",
             '--author-name', _author.name,
             '--author-email', _author.email,
             'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful', '--title', commit_message),
            ('build',),
            prepare_subprocess_mock,
            ('submit',)
        )

        assert hopic_result.exit_code == 0
        assert not expected_post_submit_commands
        if expected_version:
            repo.git.checkout('master')
            assert repo.git.describe() == expected_version


@pytest.mark.parametrize('commit_message, merge_message, expected_version, strict', (
    ('feat: initial test feature', 'feat: best feat ever', '0.1.0', True ),
    ('initial test feature'      , 'best feat ever'      , '0.1.0', False),
    ('feat: another feature'     , 'not conventional'    , '0.1.0', False),
))
def test_merge_commit_message_bump(capfd, run_hopic, commit_message, merge_message, expected_version, strict):
    result = merge_conventional_bump(capfd, run_hopic, commit_message, strict=strict, merge_message=merge_message)
    assert result.exit_code == 0


@pytest.mark.parametrize('commit_message, merge_message, expected_version, strict', (
    ('feat: a feature',       'fix: a fix',       '0.1.0', True),
))
def test_merge_commit_message_bump_error(capfd, run_hopic, commit_message, merge_message, expected_version, strict):
    result = merge_conventional_bump(capfd, run_hopic, commit_message, strict=strict, merge_message=merge_message)
    assert result.exit_code == 36


@pytest.mark.parametrize('note_mismatch', (
    False,
    True
))
def test_merge_branch_twice(run_hopic, monkeypatch, note_mismatch):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with open(run_hopic.toprepo / 'hopic-ci-config.yaml', 'w') as f:
            f.write(dedent('''\
                version:
                  bump: no
                '''))
        repo.index.add(('hopic-ci-config.yaml',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.head.reference = repo.create_head('feat/branch', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        # Some change
        with open(run_hopic.toprepo / 'something.txt', 'w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        repo.index.commit(message='feat: add something useful', **_commitargs)

    monkeypatch.setenv('GIT_COMMITTER_NAME' , 'My Name is Nobody')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'nobody@example.com')
    monkeypatch.setattr(utils, 'installed_pkgs', lambda : 'hopic==42.42.42\nhopic-dep==0.0.0\n')
    checkout_and_merge = (
        ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master', '--target-commit', str(base_commit)),
        ('prepare-source-tree',
            '--author-name', _author.name,
            '--author-email', _author.email,
            '--author-date', f"@{_git_time}",
            '--commit-date', f"@{_git_time}",
            'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'feat/branch'),
    )

    (*_, result) = run_hopic(*checkout_and_merge, ('submit',),)
    assert result.exit_code == 0

    with git.Repo(run_hopic.toprepo, expand_vars=False) as repo:
        note = repo.git.notes('show', 'master', ref='hopic/master')
        assert re.match(
                r'^Committed-by: Hopic.*\nWith Python version: .*\nAnd with these installed packages:\n.*\bhopic\b',
                note, flags=re.DOTALL | re.MULTILINE,
            )

    assert result.exit_code == 0

    if note_mismatch:
        monkeypatch.setattr(utils, 'get_package_version', lambda package: '42.42.42')

    (*_, result) = run_hopic(*checkout_and_merge)

    if note_mismatch:
        assert result.exit_code == 39
        assert result.exception is not None
    else:
        assert result.exit_code == 0
        assert result.exception is None


def test_add_hopic_config_file(run_hopic):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with open(run_hopic.toprepo / 'something.txt', 'w') as f:
            f.write('usable')
        repo.index.add(('something.txt',))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)

        # PR branch
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        with (run_hopic.toprepo / 'hopic-ci-config.yaml').open('w') as f:
            f.write(dedent('''\
                version:
                    bump: no
                '''))

        repo.index.add(('hopic-ci-config.yaml',))
        repo.index.commit(message='chore: add hopic config file', **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree', '--author-name', _author.name, '--author-email', _author.email,
                'merge-change-request', '--source-remote', run_hopic.toprepo, '--source-ref', 'something-useful'),
            ('build',),
        )
    assert result.exit_code == 0


@pytest.mark.parametrize("version_file", ("version.txt", None), ids=lambda fn: fn or "{tag}")
@pytest.mark.parametrize(
    "prepare_source_tree",
    (
        "merge-change-request",
        "apply-modality-change",
    ),
)
@pytest.mark.parametrize(
    "bump_policy",
    (
        {"policy": "constant", "field": "patch"},
        {"policy": "conventional-commits", "strict": True},
    ),
    ids=lambda bp: bp["policy"],
)
def test_hotfix_change_on_release(bump_policy, prepare_source_tree, run_hopic, version_file):
    init_version = "1.2.3"
    hotfix_id = "vindyne.mem-leak"
    expected_version = f"1.2.4-hotfix.{hotfix_id}"
    hotfix_branch = f"hotfix/{init_version}-{hotfix_id}"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                f"""\
                version:
                  tag: yes
                  format: semver
                  bump: {json.dumps(bump_policy)}
                  hotfix-branch: '^hotfix/\\d+\\.\\d+\\.\\d+-(?P<id>[a-zA-Z](?:[-.a-zA-Z0-9]*[a-zA-Z0-9])?)$'
                  hotfix-allowed-start-tags:
                    - ci
                  {("file: " + version_file) if version_file else ""}

                modality-source-preparation:
                  CHANGE:
                    - sh: touch new-file.txt
                      changed-files:
                        - new-file.txt
                      commit-message: "fix: add new file"
                """
            )
        )
        repo.index.add((cfg_file,))

        if version_file:
            (run_hopic.toprepo / version_file).write_text(f"version={init_version}")
            repo.index.add((version_file,))

        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag(init_version)
        repo.git.branch(hotfix_branch, move=True)

        if bump_policy["policy"] == "conventional-commits":
            base_commit = repo.index.commit(message="ci: prepare for hotfix", **_commitargs)

        # PR branch
        repo.head.reference = repo.create_head("fix/mem-leak", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    # Successful checkout and build
    (result,) = run_hopic(
        command("checkout-source-tree", target_remote=run_hopic.toprepo, target_ref=hotfix_branch),
    )
    if not isinstance(result.exception, (type(None), SystemExit)):
        raise result.exception
    assert result.exit_code == 0

    prepare_source_tree_params = {
        subcmd: (subcmd, *args)
        for (subcmd, *args) in (
            command(
                "merge-change-request",
                source_remote=run_hopic.toprepo,
                source_ref="fix/mem-leak",
                change_request="42",
                title="fix: work around oom kill due to memory leak",
            ),
            command(
                "apply-modality-change",
                "CHANGE",
            ),
        )
    }

    (result,) = run_hopic(
        command(
            "prepare-source-tree",
            author_name=_author.name,
            author_email=_author.email,
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
        )
        + prepare_source_tree_params[prepare_source_tree],
    )
    if not isinstance(result.exception, (type(None), SystemExit)):
        raise result.exception
    assert result.exit_code == 0
    assert result.stdout.splitlines()[-1].split("+")[0] == expected_version

    (result,) = run_hopic(
        ("submit", "--target-remote", run_hopic.toprepo),
    )
    if not isinstance(result.exception, (type(None), SystemExit)):
        raise result.exception
    assert result.exit_code == 0

    with git.Repo(run_hopic.toprepo) as repo:
        # Switch back to hotfix branch to be able to easily look at its contents
        repo.git.checkout(hotfix_branch)

        assert repo.tags[expected_version].commit == repo.head.commit


@pytest.mark.parametrize(
    "bump_policy",
    (
        {"policy": "constant", "field": "patch"},
        {"policy": "conventional-commits", "strict": True},
    ),
    ids=lambda bp: bp["policy"],
)
@pytest.mark.parametrize("unrelated_tag", (None, "1.2.4-rc1"), ids=lambda t: t or "{no-tag}")
def test_hotfix_change_off_release(bump_policy, run_hopic, unrelated_tag):
    init_version = "1.2.3"
    hotfix_id = "vindyne.mem-leak"
    hotfix_branch = f"hotfix/{init_version}-{hotfix_id}"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                f"""\
                version:
                  tag: yes
                  format: semver
                  bump: {json.dumps(bump_policy)}
                  hotfix-branch: '^hotfix/\\d+\\.\\d+\\.\\d+-(?P<id>[a-zA-Z](?:[-.a-zA-Z0-9]*[a-zA-Z0-9])?)$'
                  hotfix-allowed-start-tags:
                    - ci
                """
            )
        )
        repo.index.add((cfg_file,))

        repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag(init_version)
        repo.git.branch(hotfix_branch, move=True)

        base_commit = repo.index.commit(message="fix: unrelated cosmetic problem", **_commitargs)
        if unrelated_tag:
            repo.create_tag(unrelated_tag)

        if bump_policy["policy"] == "conventional-commits":
            base_commit = repo.index.commit(message="ci: prepare for hotfix", **_commitargs)

        # PR branch
        repo.head.reference = repo.create_head("fix/mem-leak", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", hotfix_branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "fix/mem-leak",
            "--change-request", "42", "--title", "fix: work around oom kill due to memory leak"),
    )
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert "Creating hotfixes on anything but a full release is not supported." in err


def test_hotfix_double_bump(run_hopic):
    init_version = "1.2.3"
    hotfix_id = "vindyne.mem-leak"
    expected_version = f"1.2.4-hotfix.{hotfix_id}.1"
    hotfix_branch = f"hotfix/{init_version}-{hotfix_id}"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                """\
                version:
                  tag: yes
                  format: semver
                  bump:
                    policy: conventional-commits
                    strict: yes
                  hotfix-branch: '^hotfix/\\d+\\.\\d+\\.\\d+-(?P<id>[a-zA-Z](?:[-.a-zA-Z0-9]*[a-zA-Z0-9])?)$'
                """
            )
        )
        repo.index.add((cfg_file,))

        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag(init_version)
        repo.git.branch(hotfix_branch, move=True)

        # PR branch 1
        repo.head.reference = repo.create_head("fix/mem-leak", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", hotfix_branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "fix/mem-leak",
            "--change-request", "42", "--title", "fix: work around oom kill due to memory leak"),
        ("submit",),
    )
    assert result.exit_code == 0

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        repo.git.checkout(hotfix_branch)
        repo.index.commit(message="chore: intermediate commit 1 to increase commit distance", **_commitargs)
        repo.index.commit(message="chore: intermediate commit 2 to increase commit distance", **_commitargs)
        repo.index.commit(message="chore: intermediate commit 3 to increase commit distance", **_commitargs)

        # PR branch 2
        repo.head.reference = repo.create_head("fix/out-of-bounds-access", repo.head.commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)
        repo.index.commit(message="fix: skip out of bounds read", **_commitargs)

    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", hotfix_branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "fix/out-of-bounds-access",
            "--change-request", "43", "--title", "fix: skip out of bounds read"),
    )
    assert result.exit_code == 0
    assert result.stdout.splitlines()[-1].split("+")[0] == expected_version

    (result,) = run_hopic(
        ("submit",),
    )
    assert result.exit_code == 0

    with git.Repo(run_hopic.toprepo) as repo:
        # Switch back to hotfix branch to be able to easily look at its contents
        repo.git.checkout(hotfix_branch)

        assert repo.tags[expected_version].commit == repo.head.commit


@pytest.mark.parametrize(
    "hotfix_id, error_msg",
    (
        ("42indi"  , re.compile(r"Hotfix ID '.*?' is not a valid identifier")),
        ("-42"     , re.compile(r"Hotfix ID '.*?' is not a valid identifier")),
        ("-abc"    , re.compile(r"Hotfix ID '.*?' is not a valid identifier")),
        ("abc-"    , re.compile(r"Hotfix ID '.*?' is not a valid identifier")),
        ("abc/42"  , re.compile(r"Hotfix ID '.*?' is not a valid identifier")),
        ("a"       , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("a42"     , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("a-42"    , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("a.42"    , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("a-test-1", re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("b"       , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("rc"      , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("alpha"   , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("beta"    , re.compile(r"Hotfix ID '.*?' starts with reserved prefix")),
        ("awesomeness-{init_version}-something", re.compile(r"Hotfix ID 'awesomeness-(.*?)-something' is not allowed to contain the base version '\1'")),
    ),
    ids=lambda x: x if isinstance(x, str) else "",
)
def test_hotfix_invalid_id(error_msg, hotfix_id, run_hopic):
    init_version = "1.2.3"
    hotfix_id = hotfix_id.format(init_version=init_version)
    hotfix_branch = f"hotfix/{init_version}-{hotfix_id}"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                """\
                version:
                  tag: yes
                  format: semver
                  bump:
                    policy: conventional-commits
                    strict: yes
                  hotfix-branch: '^hotfix/\\d+\\.\\d+\\.\\d+-(?P<id>.*)$'
                """
            )
        )
        repo.index.add((cfg_file,))

        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag(init_version)
        repo.git.branch(hotfix_branch, move=True)

        # PR branch
        repo.head.reference = repo.create_head("fix/mem-leak", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", hotfix_branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "fix/mem-leak",
            "--change-request", "42", "--title", "fix: work around oom kill due to memory leak"),
    )
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert error_msg.search(err)


@pytest.mark.parametrize(
    "msg_tag, error_msg",
    (
        ("refactor!", re.compile("[Bb]reaking changes are not allowed [io]n hotfix")),
        ("feat", re.compile("[Nn]ew features are not allowed [io]n hotfix")),
    ),
    ids=("breaking-change", "new-feature"),
)
def test_hotfix_rejects(error_msg, msg_tag, run_hopic):
    init_version = "1.2.3"
    hotfix_id = "vindyne"
    hotfix_branch = f"hotfix/{init_version}-{hotfix_id}"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                """\
                version:
                  tag: yes
                  format: semver
                  bump:
                    policy: conventional-commits
                    strict: yes
                  hotfix-branch: '^hotfix/\\d+\\.\\d+\\.\\d+-(?P<id>.*)$'
                """
            )
        )
        repo.index.add((cfg_file,))

        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        repo.create_tag(init_version)
        repo.git.branch(hotfix_branch, move=True)

        # PR branch
        repo.head.reference = repo.create_head("pr-42", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message=f"{msg_tag}: blorg the oompsie vatsaat", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", hotfix_branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "pr-42",
            "--change-request", "42", "--title", f"{msg_tag}: blorg the oompsie vatsaat"),
    )
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert error_msg.search(err)


@pytest.mark.parametrize("version_file", ("version.txt", None), ids=lambda fn: fn or "{tag}")
@pytest.mark.parametrize("branch_name", ("master", "hotfix/{hotfix_id}"))
def test_new_version_only(branch_name, run_hopic, monkeypatch, version_file):
    init_version = "1.2.3"
    hotfix_id = "vindyne.mem-leak"
    branch = branch_name.format(hotfix_id=hotfix_id)

    expected_build_commands = [
        ("echo", "build always"),
        ("echo", "build on new version only"),
    ]
    expected_post_submit_commands = [
        ("echo", "post submit always"),
        ("echo", "post submit on new version only"),
    ]

    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = "hopic-ci-config.yaml"

        (run_hopic.toprepo / cfg_file).write_text(
            dedent(
                f"""\
                version:
                  tag: {bool(not version_file)}
                  format: semver
                  bump:
                    policy: conventional-commits
                    strict: yes
                  hotfix-branch: '^hotfix/(?P<id>.+)$'
                  {("file: " + version_file) if version_file else ""}

                phases:
                  always:
                    pre-submit:
                      - echo "build always"
                  new-version-only-step:
                    pre-submit:
                      - run-on-change: new-version-only
                        sh: echo "build on new version only"

                post-submit:
                  always:
                    - echo "post submit always"
                  new-version-only-step:
                    - run-on-change: new-version-only
                      sh: echo "post submit on new version only"
                """
            )
        )
        repo.index.add((cfg_file,))

        if version_file:
            (run_hopic.toprepo / version_file).write_text(f"version={init_version}")
            repo.index.add((version_file,))

        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)
        if not version_file:
            repo.create_tag(init_version)
        repo.git.branch(branch, move=True)

        # PR branch
        repo.head.reference = repo.create_head("fix/mem-leak", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    def mock_check_call(expected, args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    # Successful checkout, build and submit
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", branch),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
         "--author-date", f"@{_git_time}", "--commit-date", f"@{_git_time}",
         "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "fix/mem-leak",
         "--change-request", "42", "--title", "fix: work around oom kill due to memory leak"),
        functools.partial(
            monkeypatch.setattr,
            "subprocess.check_call",
            functools.partial(mock_check_call, expected_build_commands),
        ),
        ("build",),
        functools.partial(
            monkeypatch.setattr,
            "subprocess.check_call",
            functools.partial(mock_check_call, expected_post_submit_commands),
        ),
        ("submit",),
    )

    assert result.exit_code == 0
    assert not expected_build_commands
    assert not expected_post_submit_commands


def test_no_initial_version(run_hopic):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        base_commit = repo.index.commit(message="Initial commit", **_commitargs)

        # PR branch
        repo.head.reference = repo.create_head("something-useful", base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                """\
                version:
                  format: semver
                  tag: true
                  bump:
                    policy: conventional-commits
                    strict: yes
                """
            )
        )

        repo.index.add(("hopic-ci-config.yaml",))
        repo.index.commit(message="chore: add hopic config file", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", "master"),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", "something-useful", "--title", "ci: add hopic"),
        ("build",),
    )
    assert result.exception is not None
    if not isinstance(result.exception, VersioningError):
        raise result.exception
    err = result.exception.format_message()
    assert "Failed to determine the current version while attempting to bump the version" in err


def test_merge_to_non_publishable_branch(run_hopic):
    pr_branch = "fix/mem-leak"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                """\
                version:
                  format: semver
                  tag: true
                  bump:
                    policy: conventional-commits
                    strict: yes

                publish-from-branch: 'frietjes'
                """
            )
        )
        repo.index.add(("hopic-ci-config.yaml",))
        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)

        repo.head.reference = repo.create_head(pr_branch, base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", 'master'),
        ("prepare-source-tree", "--author-name", _author.name, "--author-email", _author.email,
            "merge-change-request", "--source-remote", run_hopic.toprepo, "--source-ref", pr_branch, "--title", "chore: not interesting"),
    )
    assert result.exit_code == 0


def test_merge_with_pip_constraints(run_hopic, monkeypatch):
    constraints = "dummy==0.8.2"
    packages = ("dummy>=0.8.0",)

    pr_branch = "fix/mem-leak"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        (run_hopic.toprepo / "hopic-ci-config.yaml").write_text(
            dedent(
                f"""\
                pip:
                - packages: {json.dumps(packages)}
                """
            )
        )

        constraints_file = run_hopic.toprepo / "constraints_test.txt"
        constraints_file.write_text(constraints)

        def mock_check_call(args, *popenargs, **kwargs):
            assert args[2:5] == ["pip", "install", "-c"]
            with open(args[5]) as f:
                assert constraints in f.read()

        monkeypatch.setattr(subprocess, "check_call", mock_check_call)

        repo.index.add(("hopic-ci-config.yaml",))
        base_commit = repo.index.commit(message="chore: initial commit", **_commitargs)

        repo.head.reference = repo.create_head(pr_branch, base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        (run_hopic.toprepo / "something.txt").write_text("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="fix: work around oom kill due to memory leak", **_commitargs)

    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", "master"),
        (
            "prepare-source-tree",
            "--author-name",
            _author.name,
            "--author-email",
            _author.email,
            "--constraints",
            constraints_file.resolve(),
            "merge-change-request",
            "--source-remote",
            run_hopic.toprepo,
            "--source-ref",
            pr_branch,
            "--title",
            "chore: not interesting",
        ),
    )
    assert result.exit_code == 0


def test_modality_image_config_dir(capfd, monkeypatch, run_hopic):
    config_dir = "config_directory"
    expected_docker_args = {"docker", "--workdir=/cfg", f"--volume={run_hopic.toprepo / config_dir}:/cfg", f"--volume={run_hopic.toprepo}:/code"}
    hopic_config = dedent(
        """\
        image: buildpack-deps:18.04

        modality-source-preparation:
          DUMMY:
            - sh: pwd
            # This will fail when tests are executed in docker!
            # - sh: test -e /cfg/hopic-ci-config.yaml
            # - sh: test ! -e /code/hopic-ci-config.yaml
        """
    )

    cfg_file = f"{config_dir}/hopic-ci-config.yaml"

    old_subprocess_check_call = subprocess.check_call

    def mock_check_call(args, *popenargs, **kwargs):
        assert expected_docker_args - set(args) == set()
        expected_docker_args.clear()
        return old_subprocess_check_call(args)

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    (result,) = run_hopic(
        ("--config", cfg_file)
        + command(
            "prepare-source-tree",
            author_name=_author.name,
            author_email=_author.email,
            author_date=f"@{_git_time}",
            commit_date=f"@{_git_time}",
        )
        + command("apply-modality-change", "DUMMY"),
        config=config_file(cfg_file, hopic_config),
        rundir=run_hopic.toprepo,
    )

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert result.exit_code == 0
    assert out.splitlines()[0] == "/cfg"
    assert len(expected_docker_args) == 0



# Copyright (c) 2018 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import shlex
import subprocess
import sys
from configparser import (
    NoSectionError,
)
from pathlib import (
    PurePath,
)
from typing import (
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import git

from .execution import echo_cmd_click as echo_cmd


log = logging.getLogger(__name__)


class FatalSignal(Exception):
    def __init__(self, signum):
        self.signal = signum


class DockerContainers(object):
    """
    This context manager class manages a set of Docker containers, handling their creation and deletion.
    """
    def __init__(self):
        self.containers = set()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, tb):
        if self.containers:
            log.info('Cleaning up Docker containers: %s', ' '.join(self.containers))
            try:
                echo_cmd(subprocess.check_call, ['docker', 'rm', '-v'] + list(self.containers))
            except subprocess.CalledProcessError as e:
                log.error('Could not remove all Docker volumes, command failed with exit code %d', e.returncode)
            self.containers.clear()

    def __iter__(self):
        return iter(self.containers)

    def add(self, volume_image):
        log.info('Creating new Docker container for image %s', volume_image)
        try:
            container_id = echo_cmd(subprocess.check_output, ['docker', 'create', volume_image]).strip()
        except subprocess.CalledProcessError as e:
            log.exception('Command fatally terminated with exit code %d', e.returncode)
            sys.exit(e.returncode)

        # Container ID's consist of 64 hex characters
        if not re.match('^[0-9a-fA-F]{64}$', container_id):
            log.error('Unable to create Docker container for %s', volume_image)
            sys.exit(1)

        self.containers.add(container_id)


class HopicGitInfo(NamedTuple):
    # fmt: off
    submit_commit        : Optional[git.Commit] = None
    submit_ref           : Optional[str] = None
    submit_remote        : Optional[str] = None
    refspecs             : Sequence[str] = ()
    target_commit        : Optional[git.Commit] = None
    source_commit        : Optional[git.Commit] = None
    autosquashed_commit  : Optional[git.Commit] = None
    source_commits       : Sequence[git.Commit] = ()
    autosquashed_commits : Sequence[git.Commit] = ()
    version_bumped       : Optional[bool] = None
    # fmt: on

    @classmethod
    def from_repo(cls, repo_ctx: Union[git.Repo, str, PurePath]) -> 'HopicGitInfo':
        if isinstance(repo_ctx, git.Repo):
            repo = repo_ctx
        else:
            repo = git.Repo(repo_ctx)

        try:
            submit_ref, version_bumped, target_commit, source_commit, autosquashed_commit = None, None, None, None, None
            submit_remote = None
            refspecs: Tuple[str, ...] = ()
            source_commits: Tuple[git.Commit, ...] = ()
            autosquashed_commits: Tuple[git.Commit, ...] = ()

            submit_commit = repo.head.commit
            section = f"hopic.{submit_commit}"
            with repo.config_reader() as git_cfg:
                try:
                    # Determine remote ref for current commit
                    submit_ref = git_cfg.get(section, 'ref', fallback=None)
                    submit_remote = git_cfg.get(section, "remote", fallback=None)

                    version_bumped = git_cfg.getboolean(section, "version-bumped", fallback=None)

                    if git_cfg.has_option(section, 'refspecs'):
                        refspecs = tuple(shlex.split(git_cfg.get_value(section, 'refspecs')))

                    if git_cfg.has_option(section, 'target-commit'):
                        target_commit = repo.commit(git_cfg.get_value(section, 'target-commit'))
                    if git_cfg.has_option(section, 'source-commit'):
                        source_commit = repo.commit(git_cfg.get_value(section, 'source-commit'))
                    if git_cfg.has_option(section, 'autosquashed-commit'):
                        autosquashed_commit = repo.commit(git_cfg.get_value(section, 'autosquashed-commit'))

                    if target_commit and source_commit:
                        source_commits = tuple(git.Commit.list_items(
                            repo,
                            f"{target_commit}..{source_commit}",
                            first_parent=True,
                            no_merges=True,
                        ))
                        autosquashed_commits = source_commits
                        log.debug('Building for source commits: %s', source_commits)
                    if target_commit and autosquashed_commit:
                        autosquashed_commits = tuple(git.Commit.list_items(
                            repo,
                            f"{target_commit}..{autosquashed_commit}",
                            first_parent=True,
                            no_merges=True,
                        ))
                except NoSectionError:
                    pass
        finally:
            if not isinstance(repo_ctx, git.Repo):
                repo.close()

        return cls(
            submit_commit        = submit_commit,         # noqa: E251 "unexpected spaces around '='"
            submit_ref           = submit_ref,            # noqa: E251 "unexpected spaces around '='"
            submit_remote        = submit_remote,         # noqa: E251 "unexpected spaces around '='"
            refspecs             = refspecs,              # noqa: E251 "unexpected spaces around '='"
            source_commits       = source_commits,        # noqa: E251 "unexpected spaces around '='"
            autosquashed_commits = autosquashed_commits,  # noqa: E251 "unexpected spaces around '='"
            target_commit        = target_commit,         # noqa: E251 "unexpected spaces around '='"
            source_commit        = source_commit,         # noqa: E251 "unexpected spaces around '='"
            autosquashed_commit  = autosquashed_commit,   # noqa: E251 "unexpected spaces around '='"
            version_bumped       = version_bumped,        # noqa: E251 "unexpected spaces around '='"
        )

    @property
    def has_change(self) -> bool:
        return bool(self.refspecs)


def volume_spec_to_docker_param(volume):
    if not os.path.exists(volume['source']):
        os.makedirs(volume['source'])
    param = '{source}:{target}'.format(**volume)
    try:
        param = param + ':' + ('ro' if volume['read-only'] else 'rw')
    except KeyError:
        pass
    return param



# Copyright (c) 2018 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from pkg_resources import parse_version
import re
import subprocess
import sys
from typing import (
    Optional,
    Tuple,
    Union,
)

import click

from ..build import (
    HopicGitInfo,
)

from ..compat import metadata


def is_publish_branch(ctx, hopic_git_info=None) -> bool:
    """
    Check if the branch name is allowed to publish, if publish-from-branch is not defined in the config file, all the branches should be allowed to publish
    """

    if hopic_git_info is None:
        hopic_git_info = HopicGitInfo.from_repo(ctx.obj.workspace)
    if hopic_git_info is None or hopic_git_info.submit_ref is None:
        return False

    try:
        publish_from_branch = ctx.obj.config['publish-from-branch']
    except KeyError:
        return True

    publish_branch_pattern = re.compile(f"(?:{publish_from_branch})$")
    return publish_branch_pattern.match(re.sub("^refs/heads/", "", hopic_git_info.submit_ref)) is not None


def determine_config_file_name(ctx, workspace: Optional[Path] = None):
    """
    Determines the location of the config file, possibly falling back to a default.
    """
    try:
        return ctx.obj.config_file
    except (click.BadParameter, AttributeError):
        for fname in (
                    Path("hopic-ci-config.yaml"),
                    Path(".ci/hopic-ci-config.yaml"),
                ):
            if workspace is None:
                workspace = ctx.obj.workspace
            fname = workspace / fname
            if fname.is_file():
                return fname
        raise


def installed_pkgs():
    try:
        return subprocess.check_output((sys.executable, '-m', 'pip', 'freeze')).decode('UTF-8')
    except subprocess.CalledProcessError:
        pass


def get_package_version(package):
    """
    Consults Python's `importlib.metadata` or `importlib_metadata` package and returns the target package version.
    """
    return metadata.version(package)


def check_minimum_package_version(package: str, min_version: str) -> Tuple[bool, Union[str, None]]:
    """
    Checks whether the indicated package name is installed and meets the supplied minimum version triplet.
    Returns the result as well as the installed package version (if any, None otherwise).
    """
    try:
        version = get_package_version(package)
    except metadata.PackageNotFoundError:
        return False, None

    result = parse_version(version) >= parse_version(min_version)

    return result, version



# Copyright (c) 2019 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Sequence
import json
import os
from pathlib import Path
import re
from textwrap import dedent

import pytest

from . import config_file
from ..errors import ConfigurationError


def test_image_from_manifest(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
image: !image-from-ivy-manifest
  repository: example.com
  path: example
''',
        files={
            'dependency_manifest.xml': '''\
<?xml version="1.0" encoding="UTF-8"?>
<ivy-module version="2.0">
  <dependencies>
    <dependency name="exemplar" rev="3.1.4">
      <conf mapped="toolchain" name="default" />
    </dependency>
  </dependencies>
</ivy-module>
'''
        },
    )
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['image']['default'] == 'example.com/example/exemplar:3.1.4'


def test_image_from_cfgdir_relative_manifest(run_hopic):
    cfg_file = ".ci/some-special-config/hopic-ci-config.yaml"
    (result,) = run_hopic(
        ("--config", cfg_file, "show-config"),
        config=config_file(
            cfg_file,
            '''\
image: !image-from-ivy-manifest
  manifest: ../dependency_manifest.xml
  repository: example.com
  path: example
'''),
        files={
            '.ci/dependency_manifest.xml': '''\
<?xml version="1.0" encoding="UTF-8"?>
<ivy-module version="2.0">
  <dependencies>
    <dependency name="relative-exemplar" rev="2.7.1">
      <conf mapped="toolchain" name="default" />
    </dependency>
  </dependencies>
</ivy-module>
'''
        },
    )
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['image']['default'] == 'example.com/example/relative-exemplar:2.7.1'


def test_default_image(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
image: example
''',
    )
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['image']['default'] == 'example'


def test_default_image_type_error(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
image: yes
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': .*\bimage\b.*\bmust be\b.*\bstring\b", err, re.MULTILINE)


def test_image_type_error(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
image:
  exemplare:
    repository: example.com
    path: example
    name: exemplar
    rev: 3.1.4
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': .*\bimage\b.*\bexemplare\b.*\bmust be\b.*\bstring\b", err, re.MULTILINE)


def test_image_in_variant_type_error(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
phases:
  build:
    a:
      - image:
          exemplare:
            repository: example.com
            path: example
            name: exemplar
            rev: 3.1.4
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': .*\bimage\b.*\ba\b.*\bmust be\b.*\bstring\b", err, re.MULTILINE)


def test_bad_version_config(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
version: patch
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': .*\bversion\b.*\bmust be\b.*\bmapping\b", err, re.MULTILINE)


def test_disabled_version_bumping(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
version:
  bump: no
''',
    )

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['version']['bump']['policy'] == 'disabled'
    assert 'field' not in output['version']['bump']


def test_default_conventional_bumping(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
version:
  format: semver
  tag: 'v.{version.major}.{version.minor}.{version.patch}'
  bump:
    policy: conventional-commits
''',
    )

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['version']['bump']['policy'] == 'conventional-commits'
    assert output['version']['bump']['strict'] is False

    reject_breaking_changes_on = re.compile(output['version']['bump']['reject-breaking-changes-on'])
    reject_new_features_on = re.compile(output['version']['bump']['reject-new-features-on'])
    for major_branch in (
                'release/42',
                'rel-42',
            ):
        assert reject_breaking_changes_on.match(major_branch)
        assert not reject_new_features_on.match(major_branch)
    for minor_branch in (
                'release/42.21',
                'rel-42.21',
            ):
        assert reject_breaking_changes_on.match(minor_branch)
        assert reject_new_features_on.match(minor_branch)


def test_default_workspace_is_repo_toplevel(run_hopic):
    """This checks whether the default workspace, when a --config option is given but not a --workspace option,
    is the toplevel directory of the repository the --config file resides in."""
    cfg_file = ".ci/some-special-config/hopic-ci-config.yaml"
    (result,) = run_hopic(
        ("--config", cfg_file, "show-config"),
        config=config_file(
            cfg_file,
            '''\
volumes:
  - ${CFGDIR}:/cfg
'''),
    )

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    workspace = output['volumes']['/code']['source']
    cfgdir = output['volumes']['/cfg']['source']
    assert not cfgdir.endswith('hopic-ci-config.yaml')
    assert os.path.relpath(workspace, cfgdir) == '../..'


@pytest.mark.parametrize("cfg_file, subdir, name", (
    ("hopic-ci-config.yaml"    , "."  , "hopic-ci-config.yaml"),
    (".ci/hopic-ci-config.yaml", ".ci", "hopic-ci-config.yaml"),
))
def test_default_paths(run_hopic, cfg_file, subdir, name):
    (result,) = run_hopic(
        ("show-config",),
        config=config_file(
            cfg_file,
            dedent(
                """\
            volumes:
              - ${CFGDIR}:/cfg
                """
            ),
        ),
    )

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    workspace = Path(output["volumes"]["/code"]["source"])
    cfgdir = Path(output["volumes"]["/cfg"]["source"])
    assert cfgdir.name != name
    assert cfgdir.relative_to(workspace) == Path(subdir)


def test_default_volume_mapping_set(run_hopic):
    (result,) = run_hopic(("show-config",), config="")
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    volumes = output['volumes']

    assert set(volumes.keys()) == {"/code", "/etc/passwd", "/etc/group", "/cfg"}


def test_delete_volumes_from_default_set(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config=dedent(
            '''\
            volumes:
              - source: null
                target: /etc/passwd
              - source: null
                target: /etc/group
            '''
        ),
    )
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    volumes = output['volumes']

    assert '/etc/passwd' not in volumes
    assert '/etc/group' not in volumes


def test_disallow_phase_name_reuse(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
phases:
    a: {}
    b: {}
    a:
        x: []
        y: []
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error: [Dd]uplicate entry for key .* mapping is not permitted\b", err, re.MULTILINE)


def test_reject_sequence_in_phase(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
phases:
  a:
    - 'true'
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': phase `a`.*\bmapping\b", err, re.MULTILINE)


def test_reject_mapping_in_variant(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config='''\
phases:
  a:
    x:
      o:
        - 'true'
''',
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert re.search(r"^configuration error in '.*?\bhopic-ci-config\.yaml': variant `a.x`.*\bsequence\b", err, re.MULTILINE)


def test_devnull_config(run_hopic):
    (result,) = run_hopic(("--config", os.devnull, "show-config"))
    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output


def test_global_config_block(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config=dedent(
                                """\
                                config:
                                  version:
                                    bump: no

                                  phases:
                                    test-phase:
                                      test-variant:
                                       - echo 'bob the builder'
                                """
        ),
    )

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    assert output['version']['bump']['policy'] == 'disabled'
    assert 'field' not in output['version']['bump']
    print(output)
    assert isinstance(output['phases']['test-phase']['test-variant'], Sequence)


def test_config_is_mapping_failure(run_hopic):
    (result,) = run_hopic(
        ("show-config",),
        config=dedent(
            """\
                                  - phases:
                                    test-phase:
                                      test-variant:
                                       - echo 'bob the builder'
            """
        ),
    )
    assert isinstance(result.exception, ConfigurationError)
    assert re.search(
        r"configuration error in '.*?\bhopic-ci-config\.yaml': top level configuration should be a map, but is a list",
        result.exception.format_message(),
        re.MULTILINE,
    )


def test_config_is_mapping_empty(run_hopic):
    (result,) = run_hopic(("show-config",), config="")
    assert result.exit_code == 0



# Copyright (c) 2018 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import logging
import os
from pathlib import PurePath
import re
import sys
from typing import (
    Any,
    IO,
    Iterable,
    Iterator,
    Mapping,
    NamedTuple,
    NoReturn,
    Optional,
    Pattern,
    Tuple,
    Type,
    Union,
)

if sys.version_info[:2] >= (3, 8):
    from typing import (
        Final,
        Protocol,
    )
else:
    from typing_extensions import (
        Final,
        Protocol,
    )

from io import (
    open,
)

from .errors import VersioningError
from .types import (
    Stringable,
    TBD,
)


__all__ = (
    'CarusoVer',
    'GitVersion',
    'SemVer',
    "Version",
    'read_version',
    'replace_version',
)

log = logging.getLogger(__name__)


class _IdentifierList(tuple):
    def __str__(self) -> str:
        return '.'.join(self)


class Version(Protocol):
    version_re: Pattern  # mypy doesn't handle ClassVar
    default_tag_name: str  # mypy doesn't handle ClassVar

    major: int
    minor: int
    patch: int
    prerelease: Tuple[str, ...]

    # fmt: off
    def __iter__(self) -> Iterator[Any]: ...

    @classmethod
    def parse(cls, s: str) -> Optional["Version"]: ...

    def next_prerelease(self, seed: Union[None, str, Iterable[Stringable]] = None) -> "Version": ...
    def next_version(self, bump: Any, **kwargs: Any) -> "Version": ...
    def next_version_for_commits(self, commits: Iterable[TBD]) -> "Version": ...
    def without_meta(self) -> "Version": ...

    def __eq__(self, rhs: Any) -> bool: ...
    def __ne__(self, rhs: Any) -> bool: ...
    def __lt__(self, rhs: "Version") -> bool: ...
    def __le__(self, rhs: "Version") -> bool: ...
    def __gt__(self, rhs: "Version") -> bool: ...
    def __ge__(self, rhs: "Version") -> bool: ...
    # fmt: on


class SemVer(Version):
    """
    Semantic versioning policy.

    This policy is based on Semantic Versioning 2.0.0: https://semver.org/spec/v2.0.0.html

     * Parsing and serialization is according to the syntax specified by 'semver'.
     * Version comparison and ordering is implemented exactly as specified by 'semver'.
     * Version incrementing is implemented exactly as specified by 'semver' for the major, minor and patch fields.
      - 'semver' doesn't specify a strategy for incrementing of the prerelease field, thus:
       + given a version B, constructed by incrementing the prerelease field of A
       + and a version C, constructed by incrementing the prerelease of B
       + and a version D, constructed by incrementing the patch field of A
       + we ensure that:
        * A always sorts before B; and
        * B always sorts before C; and
        * C always sorts before D; and
        * that this relationship is transitive
    """
    __slots__ = ('major', 'minor', 'patch', 'prerelease', 'build')
    default_tag_name = "{version.major}.{version.minor}.{version.patch}{version.prerelease_separator}{version.prerelease}"

    def __init__(self, major: int, minor: int, patch: int, prerelease: Tuple[str, ...], build: Tuple[str, ...]):
        self.major      = major
        self.minor      = minor
        self.patch      = patch
        self.prerelease = prerelease
        self.build      = build

    def __setattr__(self, name, value):
        if name in {'major', 'minor', 'patch'}:
            return super().__setattr__(name, int(value))
        elif name in {'prerelease', 'build'}:
            return super().__setattr__(name, _IdentifierList(value))

    def __iter__(self) -> Iterator[TBD]:
        return iter(getattr(self, attr) for attr in self.__class__.__slots__)

    def __repr__(self) -> str:
        return '%s(major=%r, minor=%r, patch=%r, prerelease=%r, build=%r)' % ((self.__class__.__name__,) + tuple(self))

    @property
    def prerelease_separator(self) -> str:
        return "-" if self.prerelease else ""

    @property
    def build_separator(self) -> str:
        return "+" if self.build else ""

    def __str__(self) -> str:
        ver = '.'.join(str(x) for x in tuple(self)[:3])
        ver += f"{self.prerelease_separator}{self.prerelease}"
        ver += f"{self.build_separator}{self.build}"
        return ver

    version_re = re.compile(
        r'^(?:version=)?'
      + r'(?P<major>0|[1-9][0-9]*)'                                  # noqa: E131
      + r'\.(?P<minor>0|[1-9][0-9]*)'
      + r'\.(?P<patch>0|[1-9][0-9]*)'
      + r'(?:-(?P<prerelease>[-0-9a-zA-Z]+(?:\.[-0-9a-zA-Z]+)*))?'
      + r'(?:\+(?P<build>[-0-9a-zA-Z]+(?:\.[-0-9a-zA-Z]+)*))?'
      + r'\s*$'
    )

    @classmethod
    def parse(cls, s: str) -> Optional["SemVer"]:
        m = cls.version_re.match(s)
        if not m:
            return None

        major, minor, patch = (int(c) for c in m.groups()[:3])

        if m.group("prerelease") is not None:
            prerelease = tuple(m.group("prerelease").split("."))
        else:
            prerelease = ()

        if m.group("build") is not None:
            build = tuple(m.group("build").split("."))
        else:
            build = ()

        return cls(major, minor, patch, prerelease, build)

    def next_major(self) -> "SemVer":
        if self.prerelease and self.minor == 0 and self.patch == 0:
            # Just strip pre-release
            return SemVer(self.major, self.minor, self.patch, (), ())

        return SemVer(self.major + 1, 0, 0, (), ())

    def next_minor(self) -> "SemVer":
        if self.prerelease and self.patch == 0:
            # Just strip pre-release
            return SemVer(self.major, self.minor, self.patch, (), ())

        return SemVer(self.major, self.minor + 1, 0, (), ())

    def next_patch(self) -> "SemVer":
        if self.prerelease:
            # Just strip pre-release
            return SemVer(self.major, self.minor, self.patch, (), ())

        return SemVer(self.major, self.minor, self.patch + 1, (), ())

    _number_re = re.compile(r'^(?:[1-9][0-9]*|0)$')
    def next_prerelease(self, seed: Union[None, str, Iterable[Stringable]] = None) -> "SemVer":  # noqa: E301 'expected 1 blank line'
        # Special case for if we don't have a prerelease: bump patch and seed prerelease
        if not self.prerelease:
            if isinstance(seed, str):
                seed = (seed,)
            elif not seed:
                seed = ('1',)
            seed = tuple(str(i) for i in seed)

            return SemVer(self.major, self.minor, self.patch + 1, seed, ())

        # Find least significant numeric identifier to increment
        increment_idx = None
        for idx, elem in reversed(list(enumerate(self.prerelease))):
            if self._number_re.match(elem):
                increment_idx = idx
                break
        if increment_idx is None:
            return SemVer(self.major, self.minor, self.patch, self.prerelease + ('1',), ())

        # Increment only the specified identifier
        prerelease = (
            self.prerelease[:increment_idx]
          + (str(int(self.prerelease[increment_idx]) + 1),)  # noqa: E131
          + self.prerelease[increment_idx + 1:]
        )
        return SemVer(self.major, self.minor, self.patch, prerelease, ())

    def next_version(self, bump="prerelease", **kwargs: TBD) -> "SemVer":
        if bump == 'prerelease' and 'prerelease_seed' in kwargs:
            kwargs = kwargs.copy()
            kwargs['seed'] = kwargs.pop('prerelease_seed')
        return {  # type: ignore[operator]
            'prerelease': self.next_prerelease,
            'patch'     : self.next_patch,
            'minor'     : self.next_minor,
            'major'     : self.next_major,
        }[bump](**kwargs)

    def next_version_for_commits(self, commits: Iterable[TBD]) -> "SemVer":
        has_new_feature = False
        has_fix = False
        for commit in commits:
            if commit.has_breaking_change():
                return self.next_major()
            if commit.has_new_feature():
                has_new_feature = True
            if commit.has_fix():
                has_fix = True
        if has_new_feature:
            return self.next_minor()
        elif has_fix:
            return self.next_patch()
        return self

    def without_meta(self) -> "SemVer":
        return SemVer(self.major, self.minor, self.patch, self.prerelease, ())

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        if tuple(self)[:4] != tuple(rhs)[:4]:
            return False
        if self.build != rhs.build:
            return NotImplemented
        return True

    def __ne__(self, rhs: object) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        if tuple(self)[:4] != tuple(rhs)[:4]:
            return True
        if self.build != rhs.build:
            return NotImplemented
        return False

    def __lt__(self, rhs: Version) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        if tuple(self)[:3] < tuple(rhs)[:3]:
            return True
        if tuple(self)[:3] != tuple(rhs)[:3]:
            return False

        if self.prerelease and not rhs.prerelease:
            # Having a prerelease sorts before not having one
            return True
        elif not self.prerelease:
            return False

        assert self.prerelease and rhs.prerelease

        a: Union[int, str]
        b: Union[int, str]
        for a, b in zip(self.prerelease, rhs.prerelease):
            try:
                a = int(a)
            except ValueError:
                pass
            try:
                b = int(b)
            except ValueError:
                pass
            if isinstance(a, int) and not isinstance(b, int):
                # Numeric identifiers sort before non-numeric ones
                return True
            elif isinstance(a, int) != isinstance(b, int):
                return False
            if a < b:  # type: ignore[operator]
                return True
            elif b < a:  # type: ignore[operator]
                return False

        return len(self.prerelease) < len(rhs.prerelease)

    def __le__(self, rhs: Version) -> bool:
        return self < rhs or tuple(self)[:-1] == tuple(rhs)[:-1]

    def __gt__(self, rhs: Version) -> bool:
        return rhs < self

    def __ge__(self, rhs: Version) -> bool:
        return rhs <= self


class CarusoVer(Version):
    """Caruso-specific versioning policy, overlaps with semantic versioning in syntax but definitely not compatible."""
    __slots__ = ('major', 'minor', 'patch', 'prerelease', 'increment', 'fix')
    default_tag_name = "{version.major}.{version.minor}.{version.patch}{version.prerelease_separator}{version.prerelease}+PI{version.increment}.{version.fix}"

    def __init__(self, major: int, minor: int, patch: int, prerelease: Tuple[str, ...], increment: int, fix: int):
        self.major      = major
        self.minor      = minor
        self.patch      = patch
        self.prerelease = prerelease
        self.increment  = increment
        self.fix        = fix

    def __setattr__(self, name, value):
        if name in {'major', 'minor', 'patch', 'increment', 'fix'}:
            return super().__setattr__(name, int(value))
        elif name in {'prerelease'}:
            return super().__setattr__(name, _IdentifierList(value))

    def __iter__(self) -> Iterator[TBD]:
        return iter(getattr(self, attr) for attr in self.__class__.__slots__)

    def __repr__(self) -> str:
        return '%s(major=%r, minor=%r, patch=%r, prerelease=%r, increment=%r, fix=%r)' % ((self.__class__.__name__,) + tuple(self))

    @property
    def prerelease_separator(self) -> str:
        return "-" if self.prerelease else ""

    def __str__(self) -> str:
        ver = '.'.join(str(x) for x in tuple(self)[:3])
        ver += f"{self.prerelease_separator}{self.prerelease}"
        ver += f"+PI{self.increment}.{self.fix}"
        return ver

    version_re = re.compile(
        r'^(?:version=)?'
      + r'(?P<major>0|[1-9][0-9]*)'                                  # noqa: E131
      + r'\.(?P<minor>0|[1-9][0-9]*)'
      + r'\.(?P<patch>0|[1-9][0-9]*)'
      + r'(?:-(?P<prerelease>[-0-9a-zA-Z]+(?:\.[-0-9a-zA-Z]+)*))?'
      + r'\+PI(?P<increment>0|[1-9][0-9]*)\.(?P<fix>0|[1-9][0-9]*)'
      + r'\s*$'
    )

    @classmethod
    def parse(cls, s: str) -> Optional["CarusoVer"]:
        m = cls.version_re.match(s)
        if not m:
            return None

        major = int(m.group("major"))
        minor = int(m.group("minor"))
        patch = int(m.group("patch"))
        increment = int(m.group("increment"))
        fix = int(m.group("fix"))

        if m.group("prerelease") is not None:
            prerelease = tuple(m.group("prerelease").split("."))
        else:
            prerelease = ()

        return cls(major, minor, patch, prerelease, increment, fix)

    def next_fix(self) -> "CarusoVer":
        if self.prerelease:
            # Just strip pre-release
            return CarusoVer(self.major, self.minor, self.patch, (), self.increment, self.fix)

        return CarusoVer(self.major, self.minor, self.patch, (), self.increment, self.fix + 1)

    _number_re = re.compile(r'^(?:[1-9][0-9]*|0)$')
    def next_prerelease(self, seed: Union[None, str, Iterable[Stringable]] = None) -> "CarusoVer":  # noqa: E301 'expected 1 blank line'
        # Special case for if we don't have a prerelease: bump patch and seed prerelease
        if not self.prerelease:
            if isinstance(seed, str):
                seed = (seed,)
            elif not seed:
                seed = ('1',)
            seed = tuple(str(i) for i in seed)

            return CarusoVer(self.major, self.minor, self.patch, seed, self.increment, self.fix + 1)

        # Find least significant numeric identifier to increment
        increment_idx = None
        for idx, elem in reversed(list(enumerate(self.prerelease))):
            if self._number_re.match(elem):
                increment_idx = idx
                break
        if increment_idx is None:
            return CarusoVer(self.major, self.minor, self.patch, self.prerelease + ('1',), self.increment, self.fix)

        # Increment only the specified identifier
        prerelease = (
            self.prerelease[:increment_idx]
          + (str(int(self.prerelease[increment_idx]) + 1),)  # noqa: E131
          + self.prerelease[increment_idx + 1:]
        )
        return CarusoVer(self.major, self.minor, self.patch, prerelease, self.increment, self.fix)

    def next_version(self, bump="prerelease", **kwargs: TBD) -> "CarusoVer":
        if bump == 'prerelease' and 'prerelease_seed' in kwargs:
            kwargs = kwargs.copy()
            kwargs['seed'] = kwargs.pop('prerelease_seed')
        return {  # type: ignore[operator]
            'prerelease': self.next_prerelease,
            'fix'       : self.next_fix,
        }[bump](**kwargs)

    def next_version_for_commits(self, commits: Iterable[TBD]) -> NoReturn:
        raise NotImplementedError

    def without_meta(self) -> "CarusoVer":
        return self

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        return tuple(self) == tuple(rhs)

    def __ne__(self, rhs: object) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        return tuple(self) != tuple(rhs)

    def __lt__(self, rhs: Version) -> bool:
        if not isinstance(rhs, self.__class__):
            return NotImplemented
        lhs_t = tuple(self)[:3] + tuple(self)[4:6]
        rhs_t = tuple(rhs)[:3] + tuple(rhs)[4:6]
        if lhs_t < rhs_t:
            return True
        if lhs_t > rhs_t:
            return False

        if self.prerelease and not rhs.prerelease:
            # Having a prerelease sorts before not having one
            return True
        elif not self.prerelease:
            return False

        assert self.prerelease and rhs.prerelease

        a: Union[str, int]
        b: Union[str, int]
        for a, b in zip(self.prerelease, rhs.prerelease):
            try:
                a = int(a)
            except ValueError:
                pass
            try:
                b = int(b)
            except ValueError:
                pass
            if isinstance(a, int) and not isinstance(b, int):
                # Numeric identifiers sort before non-numeric ones
                return True
            elif isinstance(a, int) != isinstance(b, int):
                return False
            if a < b:  # type: ignore[operator]
                return True
            elif b < a:  # type: ignore[operator]
                return False

        return len(self.prerelease) < len(rhs.prerelease)

    def __le__(self, rhs: Version) -> bool:
        return self < rhs or self == rhs

    def __gt__(self, rhs: Version) -> bool:
        return rhs < self

    def __ge__(self, rhs: Version) -> bool:
        return rhs <= self


_rejected_hotfix_prefixes = frozenset((
    "a",
    "b",
    "c",
    "rc",
    "alpha",
    "beta",
    "pre",
    "preview",
    "post",
    "rev",
    "r",
    "dev",
))


def hotfix_id(pat: Union[str, Pattern], branch_name: Optional[str]) -> Tuple[str, ...]:
    """
    Extracts a hotfix ID from a hotfix branch name using the given regular expression.
    """
    if branch_name is None:
        return ()

    if not isinstance(pat, Pattern):
        pat = re.compile(pat)

    idx = pat.groupindex.get("id", pat.groupindex.get("ID", 1))

    m = pat.match(branch_name)
    if not m:
        return ()

    hotfix = m.group(idx)

    if not re.match(r"^[a-zA-Z](?:[-.a-zA-Z0-9]*[a-zA-Z0-9])?$", hotfix):
        raise VersioningError(f"Hotfix ID '{hotfix}' is not a valid identifier")
    prefix = re.split(r"[-.]", hotfix)[0]
    if re.sub(r"[0-9]+$", "", prefix) in _rejected_hotfix_prefixes:
        raise VersioningError(f"Hotfix ID '{hotfix}' starts with reserved prefix {prefix}")

    return _IdentifierList(hotfix.split("."))


_fmts: Mapping[str, Type[Version]] = {
    'semver': SemVer,
    'carver': CarusoVer,
}


def read_version(fname, format="semver", encoding: Optional[str] = None) -> Optional[Version]:
    fmt = _fmts[format]

    with open(fname, 'r', encoding=encoding) as f:
        for line in f:
            version = fmt.parse(line)
            if version is not None:
                return version

    return None


# NOTE: while this is a regular language, it's one who's captures cannot be described if put in a single regex
_git_describe_commit_re: Final = re.compile(r"^(?:(.*)-g)?([0-9a-f]+)$")
_git_describe_distance_re: Final = re.compile(r"^(.*)-([0-9]+)$")

_semver_tag_cleanup: Final = re.compile(r"^[^0-9]+")


class GitVersion(NamedTuple):
    tag_name     : str
    dirty        : bool = False
    commit_count : Optional[int] = None
    commit_hash  : Optional[str] = None

    @property
    def exact(self) -> bool:
        return not self.dirty and self.commit_count == 0

    @classmethod
    def from_description(cls, description: str) -> "GitVersion":
        dirty = description.endswith('-dirty')
        if dirty:
            description = description[:-len('-dirty')]

        abbrev_commit_hash = None
        commit_match = _git_describe_commit_re.match(description)
        if commit_match:
            description, abbrev_commit_hash = commit_match.groups()
            if description is None:
                description = ''

        commit_count = None
        count_match = _git_describe_distance_re.match(description)
        if count_match:
            commit_count = int(count_match.group(2))
            tag_name = count_match.group(1)
        else:
            tag_name = description

        return cls(tag_name=tag_name, dirty=dirty, commit_count=commit_count, commit_hash=abbrev_commit_hash)

    def to_version(self, format="semver", dirty_date: Optional[datetime] = None) -> Optional[Version]:
        version_part = _semver_tag_cleanup.sub("", self.tag_name)
        tag_version = _fmts[format].parse(version_part)
        if tag_version is None:
            if log.isEnabledFor(logging.WARNING):
                if self.commit_count is not None:
                    describe_out = f"{self.tag_name}-{self.commit_count}"
                else:
                    describe_out = self.tag_name
                if self.commit_hash is not None:
                    if describe_out:
                        describe_out += "-"
                    describe_out += self.commit_hash
                if describe_out and self.dirty:
                    describe_out += "-dirty"

                log.warning("Failed to parse version string %r as %s (from 'git describe' output %r)", version_part, format, describe_out)
            return None

        if self.commit_count or self.dirty:
            # Ensure that 'dirty' commits sort before the next non-dirty commit by always having a commit count
            commit_state_info = (str(self.commit_count) if self.commit_count else "0",)

            if self.dirty:
                if dirty_date is None:
                    dirty_date = datetime.utcnow()
                commit_state_info = (*commit_state_info, "dirty", f"{dirty_date:%Y%m%d%H%M%S}")

            if tag_version.prerelease:
                tag_version.prerelease = (*tag_version.prerelease, *commit_state_info)
            else:
                tag_version = tag_version.next_prerelease(seed=commit_state_info)

        if self.commit_hash is not None:
            try:
                tag_version.build = (*tag_version.build, "g" + self.commit_hash)
            except AttributeError:
                pass

        return tag_version


def replace_version(fname: PurePath, new_version: Version, encoding: Optional[str] = None, outfile: Optional[IO[str]] = None) -> None:

    out = outfile
    temp = None
    if out is None:
        out = temp = open(fname.with_suffix(fname.suffix + ".tmp"), "w", encoding=encoding)

    try:
        with open(fname, 'r', encoding=encoding) as f:
            for line in f:
                # Replace version in source line
                m = new_version.version_re.match(line)
                if m:
                    line = line[:m.start(1)] + str(new_version) + line[m.end(m.lastindex or 0):]
                out.write(line)
    except:  # noqa: E722: we re-raise, so it's not a problem
        if temp is not None:
            temp.close()
            os.remove(temp.name)
        raise
    else:
        if temp is not None:
            temp.close()
            os.rename(temp.name, fname)



# Copyright (c) 2020 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
import re
import subprocess
import sys
import tempfile
from pathlib import Path
from textwrap import dedent
from typing import (
    Optional,
)

from ..compat import metadata

import click

from ..config_reader import (
    get_entry_points,
    read as read_config,
)
from ..execution import echo_cmd_click as echo_cmd
from .utils import (
    determine_config_file_name,
    get_package_version,
    check_minimum_package_version,
)

PACKAGE : str = __package__.split('.')[0]
log = logging.getLogger(__name__)


def check_minimum_pip_version():
    """Check the installed version of pip and issue a warning if it's lower than recommended."""
    MINIMUM_PIP_VERSION = "21.1.0"
    success, version = check_minimum_package_version("pip", MINIMUM_PIP_VERSION)
    if success:
        return

    if not version:
        log.warning("Package 'pip' is not installed")
    else:
        log.warning(
            dedent(
                """
            The installed 'pip' version (%s) does not meet the recommended minimum of %s.
            You may encounter fatal pip errors related to URL constraints.
            It is strongly recommended to upgrade your pip package:

              pip install --upgrade pip>=21.0
            """
            ),
            version,
            MINIMUM_PIP_VERSION,
        )


@click.command()
# fmt: off
@click.option("--constraints", type=click.Path(exists=True, dir_okay=False, readable=True), help="""Apply the provided constraints file to pip operations""")
@click.option("--upgrade", is_flag=True, help="""Request for already installed packages to be upgraded to newer versions, can't be combined with "constraints" option""")
# fmt: on
@click.pass_context
def install_extensions(ctx, constraints: Optional[str] = None, *, upgrade: bool = False):
    if constraints and upgrade:
        raise click.BadOptionUsage("upgrade", 'options "constraints" and "upgrade" are mutually exclusive')

    def installer(pip_cfg):
        return install_extensions_with_config(pip_cfg, constraints, upgrade=upgrade)

    # Read the config file and install all templates available
    return read_config(
        determine_config_file_name(ctx),
        ctx.obj.volume_vars,
        installer,
    )


def install_extensions_with_config(pip_cfg, input_constraints_file: Optional[str], *, upgrade: bool):
    if not pip_cfg:
        return

    is_venv = (hasattr(sys, 'real_prefix') or getattr(sys, 'base_prefix', None) != sys.prefix)

    with tempfile.TemporaryDirectory() as td:
        # Prevent changing the Hopic version and add constraints that were passed in, if any
        constraints_text = f"{PACKAGE}=={get_package_version(PACKAGE)}\n"
        if input_constraints_file:
            # Issue a warning if a constraints file is provided and the pip version is not prepared
            # to handle URL constraints.
            check_minimum_pip_version()

            input_constraints = Path(input_constraints_file).read_text()
            # Remove any existing references to the Hopic package itself from the input
            constraints_text += re.sub(f"(?m)^{PACKAGE}[^A-Za-z0-9._-].*?\n", "", input_constraints)

        constraints_file = Path(td) / "constraints.txt"
        constraints_file.write_text(constraints_text)
        log.debug("pip constraints used:\n%s", constraints_text)

        base_cmd = [
            sys.executable,
            "-m",
            "pip",
            "install",
            "-c",
            str(constraints_file.resolve()),
        ]

        plog = logging.getLogger(PACKAGE)
        if plog.isEnabledFor(logging.DEBUG):
            base_cmd.append('--verbose')

        if not is_venv:
            base_cmd.append('--user')

        if upgrade:
            base_cmd.append("--upgrade")

        for spec in pip_cfg:
            cmd = base_cmd.copy()

            from_index = spec.get('from-index')
            if from_index is not None:
                cmd.extend(['--index-url', spec['from-index']])
            for index in spec['with-extra-index']:
                cmd.extend(['--extra-index-url', index])

            cmd.extend(spec['packages'])

            try:
                echo_cmd(subprocess.check_call, cmd, stdout=sys.__stderr__)
            except subprocess.CalledProcessError as exc:
                if not spec['with-extra-index']:
                    raise

                # This is the first version that fixes https://github.com/pypa/pip/issues/4195
                required_versionstr = "19.1"
                versionstr = metadata.version("pip")

                def try_int(s):
                    try:
                        return int(s)
                    except ValueError:
                        return s

                version = tuple(try_int(x) for x in versionstr.split("."))
                required_version = tuple(try_int(x) for x in required_versionstr.split("."))
                if version < required_version:
                    log.error(
                        dedent(
                            """\
                            pip failed to install with error code %i while using an extra-index.

                            The pip version available (%s) is older than %s and may contain a bug related to usage of --extra-index-url.

                            Consider updating pip to a more recent version.
                            For example: %s -m pip install --upgrade pip
                            """
                        ),
                        exc.returncode,
                        versionstr,
                        required_versionstr,
                        sys.executable,
                    )
                raise

    # Ensure newly installed packages can be imported
    importlib.invalidate_caches()
    get_entry_points.cache_clear()
